import time


class Sent:
    timestamp: int
    data: bytes
    retries: int

    def __init__(self, *, data: bytes):
        self.timestamp = int(time.time())
        self.data = data
        self.retries = 1


class Received:
    data: bytes
    drain: bool

    def __init__(self, *, data: bytes, drain: bool):
        self.data = data
        self.drain = drain



from enum import IntEnum
from typing import Optional

import pbjson
from cryptography.fernet import Fernet


class Phase(IntEnum):
    OPEN = 1
    DATA = 2
    CLOSE = 3


TOKENS = {
    'phase': 'pe',
    'ack': 'ak',
    'token': 'tn',
    'host': 'ht',
    'port': 'pt',
    'seq': 'sq',
    'data': 'da',
    'drain': 'dn',
}


class Packet:
    phase: Phase
    ack: bool
    token: Optional[str]
    host: Optional[str]
    port: Optional[int]
    seq: Optional[int]
    data: Optional[bytes]
    drain: Optional[bool]

    def __init__(
            self, 
            *, 
            phase: Phase,
            ack: bool,
            token: Optional[str] = None,
            host: Optional[str] = None,
            port: Optional[int] = None,
            seq: Optional[int] = None,
            data: Optional[bytes] = None,
            drain: Optional[bool] = None,
    ) -> None:
        self.phase = phase
        self.ack = ack
        self.token = token
        self.host = host
        self.port = port
        self.seq = seq
        self.data = data
        self.drain = drain

    @staticmethod
    async def packet(*, data: bytes, fernet: Fernet) -> 'Packet':
        json = pbjson.loads(fernet.decrypt(data))
        return Packet(
            phase=Phase(json.get(TOKENS['phase'])),
            ack=json.get(TOKENS['ack']),
            token=json.get(TOKENS['token'], None),
            host=json.get(TOKENS['host'], None),
            port=json.get(TOKENS['port'], None),
            seq=json.get(TOKENS['seq'], None),
            data=json.get(TOKENS['data'], None),
            drain=json.get(TOKENS['drain'], None),
        )

    async def binary(self, *, fernet: Fernet) -> bytes:
        return fernet.encrypt(pbjson.dumps({TOKENS[k]: v for k, v in self.__dict__.items() if v is not None}))



import re
from typing import Optional


class RawParser:
    pattern = re.compile(
        br'(?P<method>[a-zA-Z]+) (?P<uri>(\w+://)?(?P<host>[^\s\'\"<>\[\]{}|/:]+)(:(?P<port>\d+))?[^\s\'\"<>\[\]{}|]*) ')
    uri: Optional[str] = None
    host: Optional[str] = None
    port: Optional[int] = None
    method: Optional[str] = None
    error: bool = False

    def __init__(self, *, data: bytes) -> None:
        rex = self.pattern.match(data)
        if rex:
            self.uri = self.to_str(item=rex.group('uri'))
            self.host = self.to_str(item=rex.group('host'))
            self.method = self.to_str(item=rex.group('method'))
            self.port = self.to_int(item=rex.group('port'))
        else:
            self.error = True

    @staticmethod
    def to_str(*, item: Optional[bytes]) -> Optional[str]:
        if item:
            return item.decode('charmap')

    @staticmethod
    def to_int(*, item: Optional[bytes]) -> Optional[int]:
        if item:
            return int(item)

    def __str__(self) -> str:
        return str(dict(URI=self.uri, HOST=self.host, PORT=self.port, METHOD=self.method))



import asyncio
import os
from typing import Dict, Tuple
import logging

from .packet import Packet
from .telemetry import Telemetry
from .tuning import Tuning
from .link import Link


logging.basicConfig(
    format='%(asctime)s,%(msecs)03d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s',
    datefmt='%Y-%m-%d:%H:%M:%S',
    level=logging.ERROR,
)
logger = logging.getLogger(__name__)


class Proxy(asyncio.DatagramProtocol):
    transport: asyncio.DatagramTransport
    telemetry: Telemetry
    links: Dict[Tuple[str, int], Link]
    tuning: Tuning
    proxy_host: str
    proxy_port: int

    def __init__(self, *, telemetry: Telemetry, tuning: Tuning, proxy_host: str, proxy_port: int) -> None:
        self.links = dict()
        self.telemetry = telemetry
        self.tuning = tuning
        self.proxy_host = proxy_host
        self.proxy_port = proxy_port

    def connection_made(self, transport) -> None:
        self.transport = transport

    async def _datagram_received_async(self, *, data, addr) -> None:
        self.telemetry.packets_received += 1
        self.telemetry.bytes_received += len(data)
        try:
            packet = await Packet.packet(data=data, fernet=self.tuning.fernet)
        except Exception as e:
            logger.error(e)
            self.telemetry.decoding_errors += 1
            return

        link = self.links.get(addr, Link(telemetry=self.telemetry, proxy=self, addr=addr, tuning=self.tuning))
        await link.process(packet=packet)

    def datagram_received(self, data, addr) -> None:
        asyncio.create_task(self._datagram_received_async(data=data, addr=addr))

    def error_received(self, exc) -> None:
        logger.error(exc)

    def connection_lost(self, exc) -> None:
        self.transport.close()
        logger.error(exc)

    async def serve(self) -> None:
        loop = asyncio.get_event_loop()
        await loop.create_datagram_endpoint(lambda: self, local_addr=(self.proxy_host, self.proxy_port))

    async def cleanup(self) -> None:
        while True:
            await asyncio.sleep(1)
            self.telemetry.links = len(self.links)

    async def monitor(self) -> None:
        while True:
            await asyncio.sleep(1)
            os.system('clear')
            print(self.telemetry)



import datetime


class Telemetry:
    links: int = 0
    opened: int = 0
    closed: int = 0
    packets_sent: int = 0
    packets_received: int = 0
    bytes_sent: int = 0
    bytes_received: int = 0
    decoding_errors: int = 0
    processing_errors: int = 0
    token_errors: int = 0
    type_errors: int = 0
    timeout_errors: int = 0
    connection_errors: int = 0
    serving_errors: int = 0
    resending_errors: int = 0
    sent_buf_overloads: int = 0
    recv_buf_overloads: int = 0
    max_packet_size: int = 0
    resent: int = 0
    unfinished: int = 0

    def __str__(self):
        return \
            f'{datetime.datetime.now()}\n' \
            f'\tlinks: {self.links}\n' \
            f'\topened/closed: {self.opened}/{self.closed}\n' \
            f'\tpackets sent/received: {self.packets_sent}/{self.packets_received}\n' \
            f'\tbytes sent/received: {self.bytes_sent}/{self.bytes_received}\n' \
            f'\tdecoding/processing/token/type/timeout/connection/serving/resending errors: ' \
            f'{self.decoding_errors}/{self.processing_errors}/{self.token_errors}/{self.type_errors}/' \
            f'{self.timeout_errors}/{self.connection_errors}/{self.serving_errors}/{self.resending_errors}\n' \
            f'\tsent/received buffer overloads: {self.sent_buf_overloads}/{self.recv_buf_overloads}\n' \
            f'\tmax packet size: {self.max_packet_size}\n' \
            f'\tresent/unfinished: {self.resent}/{self.unfinished}'



import asyncio
import logging

from .telemetry import Telemetry
from .tuning import Tuning
from .ouija import Ouija
from .packet import Packet


logging.basicConfig(
    format='%(asctime)s,%(msecs)03d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s',
    datefmt='%Y-%m-%d:%H:%M:%S',
    level=logging.ERROR,
)
logger = logging.getLogger(__name__)


class Relay(Ouija, asyncio.DatagramProtocol):
    transport: asyncio.DatagramTransport
    proxy_host: str
    proxy_port: int

    def __init__(
            self, 
            *,
            telemetry: Telemetry,
            tuning: Tuning,
            reader: asyncio.StreamReader, 
            writer: asyncio.StreamWriter,
            proxy_host: str,
            proxy_port: int,
            remote_host: str,
            remote_port: int,
    ) -> None:
        self.telemetry = telemetry
        self.tuning = tuning
        self.reader = reader
        self.writer = writer
        self.proxy_host = proxy_host
        self.proxy_port = proxy_port
        self.remote_host = remote_host
        self.remote_port = remote_port
        self.opened = asyncio.Event()
        self.sent_buf = dict()
        self.sent_seq = 0
        self.read_closed = asyncio.Event()
        self.recv_buf = dict()
        self.recv_seq = 0
        self.write_closed = asyncio.Event()

    def connection_made(self, transport) -> None:
        self.transport = transport

    async def _datagram_received_async(self, *, data, addr) -> None:
        self.telemetry.packets_received += 1
        self.telemetry.bytes_received += len(data)
        try:
            packet = await Packet.packet(data=data, fernet=self.tuning.fernet)
        except Exception as e:
            logger.error(e)
            self.telemetry.decoding_errors += 1
            return
        await self.process(packet=packet)

    def datagram_received(self, data, addr) -> None:
        asyncio.create_task(self._datagram_received_async(data=data, addr=addr))

    def error_received(self, exc) -> None:
        logger.error(exc)

    def connection_lost(self, exc) -> None:
        asyncio.create_task(self.close())

    async def sendto(self, *, data: bytes) -> None:
        self.transport.sendto(data)

    async def on_open(self, *, packet: Packet) -> bool:
        if not packet.ack or self.opened.is_set():
            return False

        await self.write(data=b'HTTP/1.1 200 Connection Established\r\n\r\n', drain=True)
        self.opened.set()
        return True

    async def on_serve(self) -> bool:
        loop = asyncio.get_event_loop()
        await loop.create_datagram_endpoint(lambda: self, remote_addr=(self.proxy_host, self.proxy_port))

        if not await self.send_open():
            return False

        return True

    async def on_close(self) -> None:
        if not self.transport.is_closing():
            self.transport.close()



import asyncio
import logging
import time
from typing import Optional, Dict

from .telemetry import Telemetry
from .tuning import Tuning
from .primitives import Sent, Received
from .packet import Phase, Packet


logging.basicConfig(
    format='%(asctime)s,%(msecs)03d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s',
    datefmt='%Y-%m-%d:%H:%M:%S',
    level=logging.ERROR,
)
logger = logging.getLogger(__name__)


class Ouija:
    telemetry: Telemetry
    tuning: Tuning
    reader: Optional[asyncio.StreamReader]
    writer: Optional[asyncio.StreamWriter]
    remote_host: Optional[str]
    remote_port: Optional[int]
    opened: asyncio.Event
    sent_buf: Dict[int, Sent]
    sent_seq: int
    read_closed: asyncio.Event
    recv_buf: Dict[int, Received]
    recv_seq: int
    write_closed: asyncio.Event

    async def sendto(self, *, data: bytes) -> None:
        raise NotImplementedError

    async def send(self, *, data: bytes) -> None:
        await self.sendto(data=data)

        self.telemetry.packets_sent += 1
        self.telemetry.bytes_sent += len(data)
        if len(data) > self.telemetry.max_packet_size:
            self.telemetry.max_packet_size = len(data)

    async def send_retry(self, *, data: bytes, event: asyncio.Event) -> bool:
        for _ in range(self.tuning.udp_retries):
            await self.send(data=data)
            try:
                await asyncio.wait_for(event.wait(), self.tuning.udp_timeout)
            except asyncio.TimeoutError:
                self.telemetry.resent += 1
                continue
            else:
                return True
        else:
            return False

    async def read(self) -> bytes:
        return await self.reader.read(self.tuning.tcp_buffer)

    async def write(self, *, data: bytes, drain: bool) -> None:
        self.writer.write(data)
        if drain:
            await self.writer.drain()

    async def recv(self, *, seq: int, data: bytes, drain: bool) -> None:
        if seq >= self.recv_seq:
            self.recv_buf[seq] = Received(data=data, drain=drain)

        for seq in sorted(self.recv_buf.keys()):
            if seq < self.recv_seq:
                self.recv_buf.pop(seq)
            if seq == self.recv_seq:
                recv = self.recv_buf.pop(seq)
                await self.write(data=recv.data, drain=recv.drain)
                self.recv_seq += 1

        await self.send_ack_data(seq=seq)

        if len(self.recv_buf) >= self.tuning.udp_capacity:
            await self.close()
            self.telemetry.recv_buf_overloads += 1

    async def enqueue_send(self, *, data: bytes, drain: bool) -> None:
        data_packet = Packet(
            phase=Phase.DATA,
            ack=False,
            seq=self.sent_seq,
            data=data,
            drain=drain,
        )
        data_ = await data_packet.binary(fernet=self.tuning.fernet)
        self.sent_buf[self.sent_seq] = Sent(data=data_)
        await self.send(data=data_)
        self.sent_seq += 1

        if len(self.sent_buf) >= self.tuning.udp_capacity:
            await self.close()
            self.telemetry.sent_buf_overloads += 1

    async def dequeue_send(self, *, seq: int) -> None:
        self.sent_buf.pop(seq, None)

    async def send_open(self) -> bool:
        open_packet = Packet(
            phase=Phase.OPEN,
            ack=False,
            token=self.tuning.token,
            host=self.remote_host,
            port=self.remote_port,
        )
        if not await self.send_retry(
                data=await open_packet.binary(fernet=self.tuning.fernet),
                event=self.opened,
        ):
            return False

        return True

    async def send_ack_open(self) -> None:
        open_ack_packet = Packet(
            phase=Phase.OPEN,
            ack=True,
            token=self.tuning.token,
        )
        await self.send(data=await open_ack_packet.binary(fernet=self.tuning.fernet))

    async def check_token(self, *, token: str) -> bool:
        if token == self.tuning.token:
            return True

        await self.close()
        self.telemetry.token_errors += 1
        return False

    async def send_ack_data(self, *, seq: int) -> None:
        data_ack_packet = Packet(
            phase=Phase.DATA,
            ack=True,
            seq=seq,
        )
        await self.send(data=await data_ack_packet.binary(fernet=self.tuning.fernet))

    async def send_close(self) -> None:
        close_packet = Packet(
            phase=Phase.CLOSE,
            ack=False,
        )
        await self.send_retry(
            data=await close_packet.binary(fernet=self.tuning.fernet),
            event=self.read_closed,
        )

    async def send_ack_close(self) -> None:
        close_ack_packet = Packet(
            phase=Phase.CLOSE,
            ack=True,
        )
        await self.send(data=await close_ack_packet.binary(fernet=self.tuning.fernet))

    async def on_open(self, packet: Packet) -> bool:
        raise NotImplementedError

    async def process_packet(self, *, packet: Packet) -> None:
        match packet.phase:
            case Phase.OPEN:
                if not await self.check_token(token=packet.token):
                    return

                if not await self.on_open(packet=packet):
                    return

                self.telemetry.opened += 1
            case Phase.DATA:
                if not self.opened.is_set() or self.write_closed.is_set():
                    return

                if packet.ack:
                    await self.dequeue_send(seq=packet.seq)
                else:
                    await self.recv(seq=packet.seq, data=packet.data, drain=packet.drain)
            case Phase.CLOSE:
                if packet.ack:
                    self.read_closed.set()
                else:
                    await self.send_ack_close()
                    self.write_closed.set()
            case _:
                self.telemetry.type_errors += 1

    async def process(self, *, packet: Packet) -> None:
        try:
            await self.process_packet(packet=packet)
        except ConnectionError as e:
            logger.error(e)
            self.telemetry.connection_errors += 1
        except Exception as e:
            await self.close()
            logger.error(e)
            self.telemetry.processing_errors += 1

    async def resend_packets(self) -> None:
        while self.opened.is_set() or self.sent_buf:
            await asyncio.sleep(self.tuning.udp_timeout)

            for seq in sorted(self.sent_buf.keys()):
                delta = int(time.time()) - self.sent_buf[seq].timestamp

                if delta >= self.tuning.serving_timeout or self.sent_buf[seq].retries >= self.tuning.udp_retries:
                    await self.close()
                    self.telemetry.unfinished += 1
                    break

                if delta >= self.tuning.udp_timeout:
                    await self.send(data=self.sent_buf[seq].data)
                    self.sent_buf[seq].retries += 1
                    self.telemetry.resent += 1

        await self.send_close()

        try:
            await asyncio.wait_for(self.write_closed.wait(), self.tuning.serving_timeout)
        except asyncio.TimeoutError:
            pass

    async def resend(self) -> None:
        try:
            await asyncio.wait_for(self.resend_packets(), self.tuning.serving_timeout)
        except asyncio.TimeoutError:
            self.telemetry.timeout_errors += 1
        except Exception as e:
            logger.error(e)
            self.telemetry.resending_errors += 1
        finally:
            await self.close()

    async def on_serve(self) -> bool:
        raise NotImplementedError

    async def serve_stream(self) -> None:
        if not await self.on_serve():
            return

        asyncio.create_task(self.resend())

        while self.opened.is_set():
            try:
                data = await asyncio.wait_for(self.read(), self.tuning.tcp_timeout)
            except TimeoutError:
                continue

            if not data:
                break

            for idx in range(0, len(data), self.tuning.udp_payload):
                await self.enqueue_send(
                    data=data[idx:idx + self.tuning.udp_payload],
                    drain=True if len(data) - idx <= self.tuning.udp_payload else False,
                )

    async def serve(self) -> None:
        try:
            await asyncio.wait_for(self.serve_stream(), self.tuning.serving_timeout)
        except asyncio.TimeoutError:
            self.telemetry.timeout_errors += 1
        except ConnectionError as e:
            logger.error(e)
            self.telemetry.connection_errors += 1
        except Exception as e:
            logger.error(e)
            self.telemetry.serving_errors += 1

    async def on_close(self) -> None:
        raise NotImplementedError

    async def close(self) -> None:
        if self.opened.is_set():
            self.telemetry.closed += 1

        self.opened.clear()
        self.read_closed.set()
        self.write_closed.set()

        if isinstance(self.writer, asyncio.StreamWriter) and not self.writer.is_closing():
            self.writer.close()
            await self.writer.wait_closed()

        await self.on_close()



from cryptography.fernet import Fernet


class Tuning:
    fernet: Fernet
    token: str
    serving_timeout: int
    tcp_buffer: int
    tcp_timeout: int
    udp_payload: int
    udp_timeout: int
    udp_retries: int
    udp_capacity: int

    def __init__(
            self,
            *,
            fernet: Fernet,
            token: str,
            serving_timeout: int,
            tcp_buffer: int,
            tcp_timeout: int,
            udp_payload: int,
            udp_timeout: int,
            udp_retries: int,
            udp_capacity: int,
    ) -> None:
        self.fernet = fernet
        self.token = token
        self.serving_timeout = serving_timeout
        self.tcp_buffer = tcp_buffer
        self.tcp_timeout = tcp_timeout
        self.udp_payload = udp_payload
        self.udp_timeout = udp_timeout
        self.udp_retries = udp_retries
        self.udp_capacity = udp_capacity



from setuptools import setup
from codecs import open
from os import path


here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='ouija',
    version='0.2.0',
    description='Python library for building and accessing UDP-relayed TCP proxies',
    long_description=long_description,
    url='https://github.com/neurophant/ouija/',
    author='Anton Smolin',
    author_email='smolin.anton@gmail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.11',
    ],
    keywords='async https tcp udp proxy relay',
    packages=['ouija'],
    install_requires=['cryptography>=41.0.2', 'pbjson>=1.18.0'],
)
